#import "LotsaEscherView.h"
#import "GLConverter.h"

#import <sys/time.h>



@implementation LELotsaEscherView

-(id)initWithFrame:(NSRect)frame isPreview:(BOOL)preview
{
	if(self=[super initWithFrame:frame isPreview:preview useGL:YES])
	{
		[self setAnimationTimeInterval:1/60.0];
		[self setConfigName:@"ConfigSheet"];
		[self setSaverName:@"LotsaEscher" andDefaults:[NSDictionary dictionaryWithObjectsAndKeys:
			@"0",@"uSpeed",
			@"3",@"vSpeed",
			@"5151",@"transformationType",
			@"0",@"useClock",
		nil]];

		w=h=64;

		lookup=(struct texcoord *)malloc(w*h*sizeof(struct texcoord));
		vertices=(struct vertex *)malloc(w*h*sizeof(struct vertex));
    }

    return self;
}

-(void)dealloc
{
	free(lookup);
	free(vertices);

	[super dealloc];
}

-(void)startAnimationWithDefaults:(ScreenSaverDefaults *)defaults
{
	float uspeed=[defaults floatForKey:@"uSpeed"];
	float vspeed=[defaults floatForKey:@"vSpeed"];

	dudt=-uspeed*fabs(uspeed)/90.0;
	dvdt=-vspeed*fabs(vspeed)/90.0;

	u=0;
	v=-0.35;
	t=0;

	int type=[defaults integerForKey:@"transformationType"];

	int u_step=type/100-50;
	int v_step=type%100-50;

	[self precalc:u_step p2:v_step];

	[[view openGLContext] makeCurrentContext];

	glClearColor(0,0,0,0);
	glViewport(0,0,screen_w,screen_h);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glScalef(1,((float)screen_w)/((float)screen_h),1);

	glVertexPointer(2,GL_FLOAT,sizeof(struct vertex),&vertices[0].x);
	glTexCoordPointer(2,GL_FLOAT,sizeof(struct vertex),&vertices[0].u);
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);

	tex=[LEGLConverter mipMappedTexture2DFromRep:[self imageRepFromBundle:@"escher.png"]];
	if(!tex) NSLog(@"Error: LotsaEscher failed to load escher.png!");

	glBindTexture(GL_TEXTURE_2D,tex);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
	glEnable(GL_TEXTURE_2D);
	glColor3f(1,1,1);

	[self animateOneFrame];

	[NSOpenGLContext clearCurrentContext];
}

-(void)stopAnimation
{
	[[view openGLContext] makeCurrentContext];
	glDeleteTextures(1,&tex);
	[NSOpenGLContext clearCurrentContext];

	[super stopAnimation];
}

-(void)animateOneFrame
{
	[[view openGLContext] makeCurrentContext];

//	double dt=[self deltaTime];
	double dt=1.0/60.0;

	t+=dt;

	u+=dudt*dt*(1-exp(-t/10));
	v+=dvdt*dt*(1-exp(-t/10));
	u=fmod(u,1);
	v=fmod(v,1);

	for(int i=0;i<w*h;i++)
	{
		vertices[i].u=lookup[i].u+u;
		vertices[i].v=lookup[i].v+v;
	}

	for(int y=0;y<h-1;y++)
	{
		glBegin(GL_TRIANGLE_STRIP);
//		glBegin(GL_LINES);
		for(int x=0;x<w;x++)
		{
			glArrayElement(x+y*w);
			glArrayElement(x+(y+1)*w);
		}
		glEnd();
	}

	//glFlush();

	[[view openGLContext] flushBuffer];

	[NSOpenGLContext clearCurrentContext];
}

-(void)precalc:(double)p1 p2:(double)p2
{
	double u_corner=2*M_PI*p2;
	double v_corner=log(256)*p1;
	double diag=sqrt(u_corner*u_corner+v_corner*v_corner);
	double sin_a=v_corner/diag,cos_a=u_corner/diag;
	double scale=diag/2/M_PI; //u_corner;

	int i=0;
	for(int y=0;y<h;y++)
	for(int x=0;x<w;x++)
	{
		double a=((double)x)/((double)(w-1))*2*M_PI;
		double r=((double)y)/((double)(h-1))*((double)screen_w)/((double)screen_h)*1.1;

		r*=r;

		double fx=r*cos(a);
		double fy=r*sin(a);

		if(r==0) r=0.0001;

		double fu,fv;

		fu=a;
		fv=log(r);

		double tmp;


		tmp=(fu*cos_a+fv*sin_a)*scale;
		fv=(-fu*sin_a+fv*cos_a)*scale;
		fu=tmp;

		fu/=2*M_PI;
		fv/=log(256);

		lookup[i].u=fu;
		lookup[i].v=fv;

		vertices[i].x=fx;
		vertices[i].y=-fy;

		i++;
	}
}

-(void)updateConfigWindow:(NSWindow *)window usingDefaults:(ScreenSaverDefaults *)defaults
{
	[creditview setDrawsBackground:NO];

	[uslider setFloatValue:[defaults floatForKey:@"uSpeed"]];
	[vslider setFloatValue:[defaults floatForKey:@"vSpeed"]];
	[radiomatrix selectCellWithTag:[defaults integerForKey:@"transformationType"]];
}

-(void)updateDefaults:(ScreenSaverDefaults *)defaults usingConfigWindow:(NSWindow *)window
{
	[defaults setFloat:[uslider floatValue] forKey:@"uSpeed"];
	[defaults setFloat:[vslider floatValue] forKey:@"vSpeed"];
	[defaults setInteger:[[radiomatrix selectedCell] tag] forKey:@"transformationType"];
}

@end


