#import <math.h>

#define s_curve(t) (t*t*(3.-2.*t))

#ifdef PI
#undef PI
#endif
#define PI 3.1415926535897932384626433832795028841971693993751058209


#if (defined(__ppc__)||defined(__ppc64__))

#define ALTIVEC_AVAILABLE

#import <ppc_intrinsics.h>

static inline double fastfloor(double f)
{
	register double twoTo52=4503599627370496.0;
	double c=__fsel(f,-twoTo52,twoTo52);
	double result=(f-c)+c;
	if(f<result) result-=1.0;
	return result;
}

static inline float fastrsqrt(float a)
{
	return __frsqrte(a);
}

static inline float fastrsqrt1(float a)
{
	float x=__frsqrte(a);
	return 0.5f*x*(3.0f-a*x*x);
}

static inline float fastrsqrt2(float a)
{
	float x=__frsqrte(a);
	x=0.5f*x*(3.0f-a*x*x);
	return 0.5f*x*(3.0f-a*x*x);
}

static inline float fastrsqrt5(float a)
{
	float x=__frsqrte(a);
	x=0.5f*x*(3.0f-a*x*x);
	x=0.5f*x*(3.0f-a*x*x);
	x=0.5f*x*(3.0f-a*x*x);
	x=0.5f*x*(3.0f-a*x*x);
	return 0.5f*x*(3.0f-a*x*x);
}

static inline float fastsqrt(float a)
{
	return __fres(__frsqrte(a));
}

/*
If needed, you can increase the precision of the estimate returned by
frsqrte (5-bits of precision) by using the Newton-Raphson method for
improving the estimate (x0) for 1/sqrt(a) (x1 = 0.5 * x0 * [3.0 - a * x0 * x0]).
*/



#elif defined(__i386__)

#define SSE_AVAILABLE

static inline double fastfloor(double f) { return floor(f); }
static inline float fastrsqrt(float a) { return 1.0f/sqrt(a); }
static inline float fastrsqrt1(float a) { return 1.0f/sqrt(a); }
static inline float fastrsqrt2(float a) { return 1.0f/sqrt(a); }
static inline float fastrsqrt5(float a) { return 1.0f/sqrt(a); }
static inline float fastsqrt(float a) { return sqrt(a); }

#endif



static inline float fsin(float a)
{
	a/=2.0f*PI;
	float x=((a-fastfloor(a))-0.5f)*4.0f;
	float x2=x*x;
	return (((fabsf(x)-6.0f)*x2+40.0f)*x2-96.0f)*x/61.0f;
}

static inline float fcos(float a)
{
	return fsin(a+PI/2.0f);
}


