#import "LotsaGlassView.h"
#import "GLConverter.h"
#import "SphericalObject.h"
#import "FastMath.h"


@implementation LGLotsaGlassView

-(id)initWithFrame:(NSRect)frame isPreview:(BOOL)preview
{
	if(self=[super initWithFrame:frame isPreview:preview useGL:YES])
	{
		[self setAnimationTimeInterval:1/30.0];
		[self setConfigName:@"ConfigSheet"];
		[self setSaverName:@"LotsaGlass" andDefaults:[NSDictionary dictionaryWithObjectsAndKeys:
			@"2",@"detail",
			@"0.5",@"rate",
			@"0",@"clockSize",
		nil]];
    }

    return self;
}

-(void)dealloc
{
	[super dealloc];
}

-(void)startAnimationWithDefaults:(ScreenSaverDefaults *)defaults
{
//	double slow=[defaults floatForKey:@"slowMotion"];

	int detaillevel=[defaults integerForKey:@"detail"];
	double rateval=[defaults floatForKey:@"rate"];

	sgrand(time(0));

	currobj=[[LGSlicedObject alloc] initWithDetailLevel:detaillevel];

	t=0;

	currmode=0;
	nextmode=0;
	currtime=0;
	nexttime=0;
	switchtime=4+(1-rateval)*(1-rateval)*100;

	[[self openGLContext] makeCurrentContext];

//	spheretex=[[self imageRepFromBundle:@"spheremap.png"] makeTexture2D];
	refltex=[LGGLConverter mipMappedTexture2DFromRep:[self imageRepFromBundle:@"spheremap.png"]];
	refrtex=[LGGLConverter textureCubeMapFromFront:[self imageRepFromBundle:@"front.jpg"]
	back:[self imageRepFromBundle:@"back.jpg"]
	left:[self imageRepFromBundle:@"left.jpg"]
	right:[self imageRepFromBundle:@"right.jpg"]
	top:[self imageRepFromBundle:@"top.jpg"]
	bottom:[self imageRepFromBundle:@"bottom.jpg"]
	];

	glClearColor(0,0,0,0);
	glViewport(0,0,screen_w,screen_h);

	glCullFace(GL_BACK);
	glFrontFace(GL_CCW);
	glEnable(GL_CULL_FACE);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45,(float)screen_w/(float)screen_h,1,100);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glMatrixMode(GL_TEXTURE);

	glActiveTextureARB(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_CUBE_MAP,refrtex);
	glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
	glEnable(GL_TEXTURE_CUBE_MAP);

	glActiveTextureARB(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_2D,refltex);
	glEnable(GL_TEXTURE_GEN_S);
	glEnable(GL_TEXTURE_GEN_T);
	glTexGeni(GL_S,GL_TEXTURE_GEN_MODE,GL_SPHERE_MAP);
	glTexGeni(GL_T,GL_TEXTURE_GEN_MODE,GL_SPHERE_MAP);
	glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_ADD);
/*	glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_COMBINE);
	glTexEnvi(GL_TEXTURE_ENV,GL_COMBINE_RGB,GL_INTERPOLATE);
	glTexEnvi(GL_TEXTURE_ENV,GL_SOURCE0_RGB,GL_TEXTURE);
	glTexEnvi(GL_TEXTURE_ENV,GL_OPERAND0_RGB,GL_SRC_COLOR);
	glTexEnvi(GL_TEXTURE_ENV,GL_SOURCE1_RGB,GL_PREVIOUS);
	glTexEnvi(GL_TEXTURE_ENV,GL_OPERAND1_RGB,GL_SRC_COLOR);
	glTexEnvi(GL_TEXTURE_ENV,GL_SOURCE2_RGB,GL_PRIMARY_COLOR);
	glTexEnvi(GL_TEXTURE_ENV,GL_OPERAND2_RGB,GL_SRC_ALPHA);*/
	glEnable(GL_TEXTURE_2D);

	glActiveTextureARB(GL_TEXTURE0);

	glDisable(GL_BLEND);
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_DEPTH_TEST);
	glShadeModel(GL_SMOOTH);

	[self animateOneFrame];
	[self resetDeltaTime];

	[NSOpenGLContext clearCurrentContext];
}

-(void)stopAnimation
{
	[[self openGLContext] makeCurrentContext];

	[currobj release];

	glDeleteTextures(1,&refltex);
	glDeleteTextures(1,&refrtex);

	[NSOpenGLContext clearCurrentContext];

	[super stopAnimation];
}


-(void)animateOneFrame
{
	[[self openGLContext] makeCurrentContext];

	double dt=[self deltaTime];
	t+=dt;

	if(t>nexttime) [self pickNewMode];

	matrix m=matrix::translation(0,0,-5)*matrix::rotation(t*7.4/180*PI,0,0)*matrix::rotation(0,t*13/180*PI,0);

	float fade;
	if(t-currtime<4) fade=s_curve((t-currtime)/4);
	else fade=1.0f;

	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

	glColor4f(1,1,1,0.1);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glMultMatrixf(m.a);

	[currobj drawAtTime:t matrix:(matrix)m firstMode:currmode secondMode:nextmode fade:fade];

	glPopMatrix();

	glFlush();
	[[self openGLContext] flushBuffer];

	[NSOpenGLContext clearCurrentContext];
}

-(void)pickNewMode
{
	currmode=nextmode;
	currtime=nexttime;

	int count=[currobj modes]-1;
	if(count==0) nextmode=0;
	else if(count==1) nextmode=1;
	else
	{
		do { nextmode=grand()%([currobj modes]-1)+1; }
		while(nextmode==currmode);
	}

	nexttime=currtime+switchtime;
}

-(void)updateConfigWindow:(NSWindow *)window usingDefaults:(ScreenSaverDefaults *)defaults
{
	[detail setIntValue:[defaults integerForKey:@"detail"]];
	[rate setFloatValue:[defaults floatForKey:@"rate"]];
}

-(void)updateDefaults:(ScreenSaverDefaults *)defaults usingConfigWindow:(NSWindow *)window
{
	[defaults setInteger:[detail intValue] forKey:@"detail"];
	[defaults setFloat:[rate floatValue] forKey:@"rate"];
}

@end
