#import "SphericalDistortion.h"
#import "noise.h"
#import "random.h"
#import "FastMath.h"
#import "VectorMathPPC.h"




@implementation LGSphericalDistortion

-(void)calc:(int)num time:(double)t x:(float *)xarray y:(float *)yarray z:(float *)zarray a:(float *)aarray b:(float *)barray result:(float *)res
{
#ifdef ALTIVEC_AVAILABLE
	__vector float *resvec=(__vector float *)res;
	const __vector float c1=vec_const(1.0f);
	for(int i=0;i<num/4;i++) resvec[i]=c1;
#else
	for(int i=0;i<num;i++) res[i]=1.0f;
#endif
}

+(LGSphericalDistortion *)distortion
{
	return [[[self alloc] init] autorelease];
}

/*		[self deltaTime];
		for(int i=0;i<100;i++)
		[currdistortion calc:num_vertices*3 time:t x:xarray y:yarray z:zarray a:aarray b:barray result:radii];
		NSLog(@"%f",[self deltaTime]);*/

@end



@implementation LGSwallowDistortion

-(void)calc:(int)num time:(double)t x:(float *)xarray y:(float *)yarray z:(float *)zarray a:(float *)aarray b:(float *)barray result:(float *)res
{
#ifdef ALTIVEC_AVAILABLE
	__vector float *avec=(__vector float *)aarray;
	__vector float *resvec=(__vector float *)res;
//	const __vector float c0=vec_const(0.0f);
	const __vector float c0_3=vec_const(0.3f);
	const __vector float c1=vec_const(1.0f);
	const __vector float c4=vec_const(4.0f);
	const __vector float c4t=vec_const(fmod(4.0f*t,2*PI));

	for(int i=0;i<num/4;i++)
	{
		__vector float a=avec[i];
		resvec[i]=vec_madd(c0_3,vec_sin(vec_madd(c4,a,c4t)),c1);
	}
#else
	const float c4t=fmod(4.0*t,2*PI);
	for(int i=0;i<num;i++)
	{
		float a=aarray[i];
		res[i]=1.0f+0.3f*fsin(4.0f*a+c4t);
	}
#endif
}

@end



@implementation LGFreakazoidDistortion

-(void)calc:(int)num time:(double)t x:(float *)xarray y:(float *)yarray z:(float *)zarray a:(float *)aarray b:(float *)barray result:(float *)res
{
#ifdef ALTIVEC_AVAILABLE
	__vector float *avec=(__vector float *)aarray;
	__vector float *bvec=(__vector float *)barray;
	__vector float *resvec=(__vector float *)res;
//	const __vector float c0=vec_const(0.0f);
	const __vector float c0_1=vec_const(0.1f);
	const __vector float c1=vec_const(1.0f);
	const __vector float c6=vec_const(6.0f);
	const __vector float c8=vec_const(8.0f);
	const __vector float c4t=vec_const(fmod(4.0f*t,2*PI));

	for(int i=0;i<num/4;i++)
	{
		__vector float a=avec[i],b=bvec[i];

		resvec[i]=vec_madd(
			vec_sin(a),
			c0_1*vec_sin(vec_madd(c6,b,c8*a)),
			vec_madd(c0_1,vec_sin(vec_madd(c8,a,c4t)),c1)
		);
	}
#else
	const float c4t=fmod(4.0*t,2*PI);
	for(int i=0;i<num;i++)
	{
		float a=aarray[i],b=barray[i];
		res[i]=1.0f+0.1f*fsin(8.0f*a+c4t)+0.1f*fsin(a)*fsin(6.0f*b+8.0f*a);
	}
#endif
}

@end



@implementation LGGlassNibDistortion

-(void)calc:(int)num time:(double)t x:(float *)xarray y:(float *)yarray z:(float *)zarray a:(float *)aarray b:(float *)barray result:(float *)res
{
#ifdef ALTIVEC_AVAILABLE
	__vector float *avec=(__vector float *)aarray;
	__vector float *bvec=(__vector float *)barray;
	__vector float *resvec=(__vector float *)res;
	const __vector float c0=vec_const(0.0f);
	const __vector float c0_05=vec_const(0.05f);
	const __vector float c0_1=vec_const(0.1f);
	const __vector float cm1=vec_const(-1.0f);
	const __vector float c1_5=vec_const(1.5f);
	const __vector float c10=vec_const(10.0f);
	const __vector float c4t=vec_const(fmod(4.0f*t,2*PI));
	const __vector float c4sin0_3t=vec_const(4.0f*sin(0.3*t));

	for(int i=0;i<num/4;i++)
	{
		__vector float a=avec[i],b=bvec[i];

		resvec[i]=vec_madd(
			vec_sin(a),
			vec_madd(c0_1,vec_sin(vec_madd(c10,b,vec_madd(a,c4sin0_3t,c0))),cm1),
			vec_madd(c0_05,vec_sin(vec_madd(c10,a,c4t)),c1_5)
		);
	}
#else
	const float c4t=fmod(4.0*t,2*PI);
	const float c4sin0_3t=4.0f*fsin(0.3f*t);
	for(int i=0;i<num;i++)
	{
		float a=aarray[i],b=barray[i];
		res[i]=1.5f+fsin(a)*(-1.0f+0.1f*fsin(10.0f*b+c4sin0_3t*a))+0.05f*fsin(10.0f*a+c4t);
	}
#endif
}

@end


@implementation LGPotteryDistortion

-(id)init
{
	if(self=[super init])
	{
		v1=drand()*1+0.75;
		v2=drand()*1+0.75;
	}
	return self;
}

-(void)calc:(int)num time:(double)t x:(float *)xarray y:(float *)yarray z:(float *)zarray a:(float *)aarray b:(float *)barray result:(float *)res
{
#ifdef ALTIVEC_AVAILABLE
	__vector float *avec=(__vector float *)aarray;
	__vector float *resvec=(__vector float *)res;
	const __vector float c0_3=vec_const(0.3f);
	const __vector float c1=vec_const(1.0f);
	const __vector float c4=vec_const(4.0f);
	const __vector float c6=vec_const(6.0f);
	const __vector float cv1t=vec_const(fmod(v1*t,2*PI));
	const __vector float cv2t=vec_const(fmod(v2*t,2*PI));

	for(int i=0;i<num/4;i++)
	{
		__vector float a=avec[i];
		__vector float fsa1=vec_sin(vec_madd(c4,a,cv1t));
		__vector float fsa2=vec_sin(vec_madd(c6,a,cv2t));
		fsa1*=fsa1;
		fsa2*=fsa2;

		resvec[i]=vec_madd(c0_3,fsa1*fsa2,c1);
	}
#else
	const float cv1t=fmod(v1*t,2*PI);
	const float cv2t=fmod(v2*t,2*PI);
	for(int i=0;i<num;i++)
	{
		float a=aarray[i];
		float fsa1=fsin(4*a+cv1t);
		float fsa2=fsin(6*a+cv2t);
		res[i]=1.0f+0.3f*fsa1*fsa1*fsa2*fsa2;
	}
#endif
}

@end


@implementation LGTurbineDistortion

-(void)calc:(int)num time:(double)t x:(float *)xarray y:(float *)yarray z:(float *)zarray a:(float *)aarray b:(float *)barray result:(float *)res
{
#ifdef ALTIVEC_AVAILABLE
	__vector float *avec=(__vector float *)aarray;
	__vector float *bvec=(__vector float *)barray;
	__vector float *resvec=(__vector float *)res;
	const __vector float c0_3=vec_const(0.3f);
	const __vector float c0_8=vec_const(0.8f);
	const __vector float c2=vec_const(2.0f);
	const __vector float c4=vec_const(4.0f);
	const __vector float c6=vec_const(6.0f);
	const __vector float c16t=vec_const(fmod(16.0*t,2*PI));

	for(int i=0;i<num/4;i++)
	{
		__vector float a=avec[i],b=bvec[i];
		__vector float d1=c4*vec_cos(a);
		d1*=d1; d1*=d1; d1*=d1;
		__vector float d2=c0_8*vec_sin(a);
		d2*=d2; d2*=d2;
		__vector float offs=vec_fastrsqrt2(vec_fastrsqrt2(vec_fastrsqrt2(vec_madd(d2,d2,d1))));
		__vector float sn1=vec_sin(vec_madd(c6,b,c16t));
		__vector float sn2=vec_sin(c2*a);
		__vector float sn=sn1*sn2*sn2;
		resvec[i]=vec_madd(c0_3,sn*sn,offs);
	}
#else
	const float c16t=fmod(16.0*t,2*PI);
	for(int i=0;i<num;i++)
	{
		float a=aarray[i],b=barray[i];
		float d1=4.0f*fcos(a);
		d1*=d1; d1*=d1; d1*=d1;
		float d2=0.8f*fsin(a);
		d2*=d2; d2*=d2; d2*=d2;
		float offs=fastrsqrt2(fastrsqrt2(fastrsqrt2(d1+d2)));
		float sn1=fsin(6.0f*b+c16t);
		float sn2=fsin(2.0f*a);
		float sn=sn1*sn2*sn2;
		res[i]=offs+0.3f*sn*sn;
	}
#endif
}

@end



@implementation LGCubesDistortion

-(void)calc:(int)num time:(double)t x:(float *)xarray y:(float *)yarray z:(float *)zarray a:(float *)aarray b:(float *)barray result:(float *)res
{
#ifdef ALTIVEC_AVAILABLE
	__vector float *avec=(__vector float *)aarray;
	__vector float *bvec=(__vector float *)barray;
	__vector float *resvec=(__vector float *)res;
	const __vector float c0_16=vec_const(0.16f);
	const __vector float c0_3=vec_const(0.3f);
	const __vector float c0_7=vec_const(0.7f);
	const __vector float c3=vec_const(3.0f);
	const __vector float c4=vec_const(4.0f);
	const __vector float c12=vec_const(12.0f);
	const __vector float ct=vec_const(fmod(t,2*PI));

	for(int i=0;i<num/4;i++)
	{
		__vector float a=avec[i],b=bvec[i];
		__vector float v4a_t=vec_madd(c4,a,ct);
		__vector float sn1=vec_madd(c0_16,vec_sin(c3*v4a_t),vec_sin(v4a_t));
		__vector float sn2=vec_madd(c0_16,vec_sin(c12*b),vec_sin(c4*b));
		__vector float sn=sn1*sn2;
		__vector float sqrtsn=vec_fastrsqrt1(vec_fastrsqrt1(vec_sin(a)));

		resvec[i]=vec_madd(c0_3,vec_madd(sqrtsn,sn,sqrtsn),c0_7);
	}
#else
	float ct=fmod(t,2*PI);
	for(int i=0;i<num;i++)
	{
		float a=aarray[i],b=barray[i];
		float v4a_t=4.0f*a+ct;
		float sn1=fsin(v4a_t)+0.16f*fsin(3.0f*v4a_t);
		float sn2=fsin(4.0f*b)+0.16f*fsin(12.0f*b);
		float sn=sn1*sn2;
		res[i]=0.7f+0.3f*(1.0f+sn)*fastrsqrt1(fastrsqrt1(fsin(a)));
	}
#endif
}

@end




@implementation LGSin3Distortion

-(id)init
{
	if(self=[super init])
	{
		vx=drand()*1+0.75;
		vy=drand()*1+0.75;
		vz=drand()*1+0.75;
	}
	return self;
}

-(void)calc:(int)num time:(double)t x:(float *)xarray y:(float *)yarray z:(float *)zarray a:(float *)aarray b:(float *)barray result:(float *)res
{
#ifdef ALTIVEC_AVAILABLE
	__vector float *xvec=(__vector float *)xarray;
	__vector float *yvec=(__vector float *)yarray;
	__vector float *zvec=(__vector float *)zarray;
	__vector float *resvec=(__vector float *)res;
	const __vector float c0_2=vec_const(0.2f);
	const __vector float c3pid2=vec_const(3.0f*PI/2.0f);
	const __vector float c1=vec_const(1.0f);
	const __vector float cvxt=vec_const(fmod(vx*t,2*PI));
	const __vector float cvyt=vec_const(fmod(vy*t,2*PI));
	const __vector float cvzt=vec_const(fmod(vz*t,2*PI));

	for(int i=0;i<num/4;i++)
	{
		__vector float x=xvec[i],y=yvec[i],z=zvec[i];

		resvec[i]=vec_madd(c0_2,
			vec_sin(vec_madd(c3pid2,x,cvxt))+
			vec_sin(vec_madd(c3pid2,y,cvyt))+
			vec_sin(vec_madd(c3pid2,z,cvzt)),
		c1);
	}
#else
	const float cvxt=fmod(vx*t,2*PI);
	const float cvyt=fmod(vy*t,2*PI);
	const float cvzt=fmod(vz*t,2*PI);
	for(int i=0;i<num;i++)
	{
		res[i]=1.0f+0.2f*(
			fsin(3.0f*xarray[i]*PI/2.0f+cvxt)
			+fsin(3.0f*yarray[i]*PI/2.0f+cvyt)
			+fsin(3.0f*zarray[i]*PI/2.0f+cvzt)
		);
	}
#endif
}

@end



@implementation LGSin5Distortion

-(id)init
{
	if(self=[super init])
	{
		vx=drand()*1+0.75;
		vy=drand()*1+0.75;
		vz=drand()*1+0.75;
	}
	return self;
}

-(void)calc:(int)num time:(double)t x:(float *)xarray y:(float *)yarray z:(float *)zarray a:(float *)aarray b:(float *)barray result:(float *)res
{
#ifdef ALTIVEC_AVAILABLE
	__vector float *xvec=(__vector float *)xarray;
	__vector float *yvec=(__vector float *)yarray;
	__vector float *zvec=(__vector float *)zarray;
	__vector float *resvec=(__vector float *)res;
	const __vector float c0_2=vec_const(0.2f);
	const __vector float c5pid2=vec_const(5.0f*PI/2.0f);
	const __vector float c1=vec_const(1.0f);
	const __vector float cvxt=vec_const(fmod(vx*t,2*PI));
	const __vector float cvyt=vec_const(fmod(vy*t,2*PI));
	const __vector float cvzt=vec_const(fmod(vz*t,2*PI));

	for(int i=0;i<num/4;i++)
	{
		__vector float x=xvec[i],y=yvec[i],z=zvec[i];

		resvec[i]=vec_madd(c0_2,
			vec_sin(vec_madd(c5pid2,x,cvxt))+
			vec_sin(vec_madd(c5pid2,y,cvyt))+
			vec_sin(vec_madd(c5pid2,z,cvzt)),
		c1);
	}
#else
	const float cvxt=fmod(vx*t,2*PI);
	const float cvyt=fmod(vy*t,2*PI);
	const float cvzt=fmod(vz*t,2*PI);
	for(int i=0;i<num;i++)
	{
		res[i]=1.0f+0.2f*(
			fsin(5.0f*xarray[i]*PI/2.0f+cvxt)
			+fsin(5.0f*yarray[i]*PI/2.0f+cvyt)
			+fsin(5.0f*zarray[i]*PI/2.0f+cvzt)
		);
	}
#endif
}

@end



@implementation LGSuperQuadricDistortion

-(void)calc:(int)num time:(double)t x:(float *)xarray y:(float *)yarray z:(float *)zarray a:(float *)aarray b:(float *)barray result:(float *)res
{
	double e=2*exp(sin(t));
	double inv_e=-1.0/e;
	for(int i=0;i<num;i++)
	{
		res[i]=pow(
			pow(fabs(xarray[i]),e)+
			pow(fabs(yarray[i]),e)+
			pow(fabs(zarray[i]),e
		),inv_e);
	}
}

@end

@implementation LGExpPerlinDistortion

-(void)calc:(int)num time:(double)t x:(float *)xarray y:(float *)yarray z:(float *)zarray a:(float *)aarray b:(float *)barray result:(float *)res
{
	for(int i=0;i<num;i++)
	{
		res[i]=1.0f+0.03f*(exp(10.0*perlin_noise_4d(xarray[i]+0.1*t+12.7,yarray[i]+3.14,zarray[i]-7.12,0.1*t,2.0,2.0,2))-1.0);
	}
}

@end



@implementation LGTestDistortion

-(void)calc:(int)num time:(double)t x:(float *)xarray y:(float *)yarray z:(float *)zarray a:(float *)aarray b:(float *)barray result:(float *)res
{
	for(int i=0;i<num;i++)
	{
		double a=aarray[i],b=barray[i];
		double x=xarray[i],y=yarray[i],z=zarray[i];
//		res[i]=1.0f;
//		res[i]=1.0f+0.1f*fsin(8*a+4*t)+0.1*fsin(a)*fsin(8*b);
//		res[i]=1.0f+0.005f*fsin(10*a+4*t);
//		res[i]=exp(0.4*fsin(4*a+4*t));
//		res[i]=1.0f+1.0f*perlin_noise_4d(xarray[i]+0.1f*t+12.7,yarray[i]+3.14,zarray[i]-7.12,0.1f*t,2.0,2.0,2);
//		res[i]=1.0f-0.5f*fsin(4*b)*fsin(4*b)*fsin(4*a+t)*fsin(4*a+t)*fsin(a);

/*		const float sa=fsin(3*a);
		const float sb=fcos(3*b+sa*fsin(t));
		res[i]=0.5f+0.05f*exp(3.0*sa*sa*sb*sb);*/

//		res[i]=1.0f+0.05*fsin(8*(a+2*b));

/*		float ap=floor(a*8.0f/2.0f/PI);
		ap=(ap+0.5f)*2.0f*PI/8.0f;
		float bp=floor(b*6.0f/2.0f/PI);
		bp=(bp+0.5f)*2.0f*PI/6.0f;
		float px=fsin(ap)*fcos(bp);
		float py=fsin(ap)*fsin(bp);
		float pz=fcos(ap);
		res[i]=(1/(px*x+py*y+pz*z));*/

/*		int rnd=t;
		rnd*=69069; rnd>>=8; rnd*=69069;
		float n=rnd%4+4;
		float fade=t-floor(t);
		fade=2*fade-1;
		fade=(2*fabs(fade)-3)*fade*fade+1;

		float sn=fade*fsin(n*b);

		float d1=2.0f*fcos(a);
		d1*=d1; d1*=d1; d1*=d1;
		float d2=(1.0f-0.25f*sn)*fsin(a);
		d2*=d2; d2*=d2; d2*=d2;
		float offs=fastrsqrt2(fastrsqrt2(fastrsqrt2(d1+d2)));
//		res[i]=offs+0.05f*fsin(offs*z+4*t)*fsin(offs*+4*t);
		res[i]=offs;*/

		res[i]=1;
	}
}

@end
