#import "SphericalObject.h"
#import "FastMath.h"



@implementation LGSphericalObject

-(id)initWithVertices:(int)nvertices andStrips:(int)nstrips ofSize:(int)stripsum
{
	if(self=[super initWithVertices:nvertices andStrips:nstrips ofSize:stripsum])
	{
		xarray=(float *)malloc(sizeof(float)*num_vertices*3);
		yarray=(float *)malloc(sizeof(float)*num_vertices*3);
		zarray=(float *)malloc(sizeof(float)*num_vertices*3);
		aarray=(float *)malloc(sizeof(float)*num_vertices*3);
		barray=(float *)malloc(sizeof(float)*num_vertices*3);
		radii=(float *)malloc(sizeof(float)*num_vertices*3);
		otherradii=(float *)malloc(sizeof(float)*num_vertices*3);
	}
	return self;
}

-(void)dealloc
{
	free(xarray);
	free(yarray);
	free(zarray);
	free(aarray);
	free(barray);
	free(radii);
	free(otherradii);

	[super dealloc];
}

-(void)drawAtTime:(double)t matrix:(matrix)m firstMode:(int)mode1 secondMode:(int)mode2 fade:(float)fade
{
	if(fade==0.0f)
	{
		[[distortions objectAtIndex:mode1] calc:num_vertices*3
		time:t x:xarray y:yarray z:zarray a:aarray b:barray result:radii];
	}
	else if(fade==1.0f)
	{
		[[distortions objectAtIndex:mode2] calc:num_vertices*3
		time:t x:xarray y:yarray z:zarray a:aarray b:barray result:radii];
	}
	else
	{
		[[distortions objectAtIndex:mode1] calc:num_vertices*3
		time:t x:xarray y:yarray z:zarray a:aarray b:barray result:radii];
		[[distortions objectAtIndex:mode2] calc:num_vertices*3
		time:t x:xarray y:yarray z:zarray a:aarray b:barray result:otherradii];
		[self calcFade:fade];
	}

//	vector light=vector(-1,1,1)/sqrt(3);
	vector cam=m.fast_inverse().position();

	for(int i=0;i<num_vertices;i++)
	{
		vector v=vector(xarray[i*3],yarray[i*3],zarray[i*3])*radii[i*3];

		vector n;
		if(i==num_vertices-2) n=vector(0,0,1);
		else if(i==num_vertices-1) n=vector(0,0,-1);
		else
		{
			vector v2=vector(xarray[i*3+1],yarray[i*3+1],zarray[i*3+1])*radii[3*i+1];
			vector v3=vector(xarray[i*3+2],yarray[i*3+2],zarray[i*3+2])*radii[3*i+2];

			n=(v2-v)*(v3-v);
			//n.normalize();
			n*=fastrsqrt(n.abs2());
		}

//		float c=n.dot(light);
//		uint8 cb=c>0.0f?((int)(c*255.0f)):0;
//		uint8 cb=0xff;

		vector vv=v-cam;
		//vv.normalize();
		vv*=fastrsqrt(vv.abs2());

		float n1=1.0f,n2=2.5f;
		float cos_a=fabs(n.dot(vv));
		//float cos_b=sqrt(1.0f-(n1*n1/n2/n2)*(1.0f-cos_a*cos_a));
		float cos_b=fastsqrt(1.0f-(n1*n1/(n2*n2))*(1.0f-cos_a*cos_a));

		float r12s=(n1*cos_a-n2*cos_b)/(n1*cos_a+n2*cos_b);
		float r12p=(n2*cos_a-n1*cos_b)/(n2*cos_a+n1*cos_b);
		float R=(r12s*r12s+r12p*r12p)/2.0f;

		uint8 trans=255-(int)(R*255.0f);

		vector refr=vv-n*(n2/n1*cos_b-cos_a);
		//refr.normalize();
		refr*=fastrsqrt(refr.abs2());

		vertices[i].v=v;
		vertices[i].n=n;
/*		vertices[i].r=cb;
		vertices[i].g=cb;
		vertices[i].b=cb;
		vertices[i].a=al;*/
		vertices[i].r=trans;
		vertices[i].g=trans;
		vertices[i].b=trans;
		vertices[i].a=255;
		vertices[i].tex=refr;
	}

	[super drawStrips];
}

-(void)calcFade:(float)fade
{
#ifdef USE_ALTIVEC
	__vector float *radiivec=(__vector float *)radii;
	__vector float *othervec=(__vector float *)otherradii;
	const __vector float c0=vec_const(0);
	const __vector float cfade=vec_const(fade);
	const __vector float c1mfade=vec_const(1-fade);
	for(int i=0;i<(num_vertices*3)/4;i++)
	radiivec[i]=vec_madd(radiivec[i],c1mfade,vec_madd(othervec[i],cfade,c0));
#else
	for(int i=0;i<num_vertices*3;i++)
	radii[i]=radii[i]*(1-fade)+otherradii[i]*fade;
#endif
}

-(void)setDistortions:(NSArray *)distarray;
{
	[distortions autorelease];
	distortions=[distarray retain];
}

-(int)modes
{
	return [distortions count];
}

@end



@implementation LGSlicedObject

#define DELTA 0.0003

-(id)initWithDetailLevel:(int)level
{
	int n=64+32*level;
	int m=n;

	int nvertices=(n-1)*m+2;
	int nstrips=n;
	int stripsum=2*n*(m+2);

	if(self=[super initWithVertices:nvertices andStrips:nstrips ofSize:stripsum])
	{
		int curr=0;
		for(int i=1;i<n;i++)
		{
			float a=(float)i/(float)n*PI;

			for(int j=0;j<m;j++)
			{
				float b=(float)j/(float)m*2*PI;

				xarray[curr]=sin(a)*cos(b);
				yarray[curr]=sin(a)*sin(b);
				zarray[curr]=cos(a);
				aarray[curr]=a;
				barray[curr]=b;
				curr++;

				xarray[curr]=sin(a+DELTA)*cos(b);
				yarray[curr]=sin(a+DELTA)*sin(b);
				zarray[curr]=cos(a+DELTA);
				aarray[curr]=a+DELTA;
				barray[curr]=b;
				curr++;

				xarray[curr]=sin(a+DELTA)*cos(b+DELTA);
				yarray[curr]=sin(a+DELTA)*sin(b+DELTA);
				zarray[curr]=cos(a+DELTA);
				aarray[curr]=a+DELTA;
				barray[curr]=b+DELTA;
				curr++;
			}
		}

		xarray[curr]=0;
		yarray[curr]=0;
		zarray[curr]=1;
		aarray[curr]=0;
		barray[curr]=0;
		curr+=3;

		xarray[curr]=0;
		yarray[curr]=0;
		zarray[curr]=-1;
		aarray[curr]=PI;
		barray[curr]=0;
		curr+=3;

		int *sptr=strips;

		for(int i=0;i<n;i++)
		{
			*sptr++=2*m+2;
			for(int j=0;j<m;j++)
			{
				if(i==0) *sptr++=num_vertices-2;
				else *sptr++=(i-1)*m+j;

				if(i==n-1) *sptr++=num_vertices-1;
				else *sptr++=i*m+j;
			}

			if(i==0) *sptr++=num_vertices-2;
			else *sptr++=(i-1)*m;

			if(i==n-1) *sptr++=num_vertices-1;
			else *sptr++=i*m;
		}

		[self setDistortions:[NSArray arrayWithObjects:
			[LGSphericalDistortion distortion],
//			[LGTestDistortion distortion],
			[LGSwallowDistortion distortion],
			[LGFreakazoidDistortion distortion],
			[LGGlassNibDistortion distortion],
			[LGPotteryDistortion distortion],
			[LGTurbineDistortion distortion],
			[LGCubesDistortion distortion],
//			[LGSin3Distortion distortion],
			[LGSin5Distortion distortion],
		nil]];
	}
	return self;
}

@end

