#ifdef ALTIVEC_AVAILABLE

#import <altivec.h>
#undef vector

#define vec_const(a) {(a),(a),(a),(a)}



static inline __vector float vec_sin(__vector float a)
{
	const __vector float ci2pi=vec_const(1.0f/2.0f/PI);
	const __vector float c2=vec_const(2.0f);
	const __vector float c4=c2*c2;
	const __vector float c6=c4+c2;
	const __vector float c40=vec_madd(c6,c6,c4);
	const __vector float cm96=vec_const(-96.0f);
	const __vector float ci61=vec_const(1.0f/61.0f);

	a*=ci2pi;
	__vector float x=vec_madd(a-vec_ceil(a),c4,c2);
	__vector float x2=x*x;
	__vector float res=vec_madd(x2,vec_madd(x2,vec_abs(x)-c6,c40),cm96)*x*ci61;

	return res;
}

static inline __vector float vec_cos(__vector float a)
{
	const __vector float cpid2=vec_const(PI/2.0f);
	return vec_sin(a+cpid2);
}

static inline __vector float vec_fastrsqrt1(__vector float a)
{
    __vector float c0=vec_const(0);
    __vector float c0_5=vec_const(0.5);
    __vector float c1=vec_const(1.0);
    __vector float x=vec_rsqrte(a);

    __vector float x2=vec_madd(x,x,c0);
    __vector float xd2=vec_madd(x,c0_5,c0);
	return vec_madd(vec_nmsub(a,x2,c1),xd2,x);
} 

static inline __vector float vec_fastrsqrt2(__vector float a)
{
	__vector float c0=vec_const(0);
	__vector float c0_5=vec_const(0.5);
	__vector float c1=vec_const(1.0);
	__vector float x=vec_rsqrte(a);

	__vector float x2=vec_madd(x,x,c0);
	__vector float xd2=vec_madd(x,c0_5,c0);
	x=vec_madd(vec_nmsub(a,x2,c1),xd2,x);
	x2=vec_madd(x,x,c0);
	xd2=vec_madd(x,c0_5,c0);
	return vec_madd(vec_nmsub(a,x2,c1),xd2,x);
}

#endif
