#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "noise.h"
#include "random.h"



#define B 0x100
#define BM 0xff
#define N 0x1000
#define NP 12   /* 2^N */
#define NM 0xfff

#define s_curve(t) (t*t*(3.-2.*t))

#define lerp(t,a,b) (a+t*(b-a))

#define setup(i,b0,b1,r0,r1)\
	t=vec[i]+N;\
	b0=((int)t)&BM;\
	b1=(b0+1)&BM;\
	r0=t-((int)t);\
	r1=r0-1.;

#define at2(rx,ry) (rx*q[0]+ry*q[1])
#define at3(rx,ry,rz) (rx*q[0]+ry*q[1]+rz*q[2])
#define at4(rx,ry,rz,rt) (rx*q[0]+ry*q[1]+rz*q[2]+rt*q[3])



static int p[2*B+2];
static double g4[2*B+2][4];
static double g3[2*B+2][3];
static double g2[2*B+2][2];
static double g1[2*B+2];
static int start=1;



double noise1(double arg)
{
	int bx0,bx1;
	double rx0,rx1,sx,t,u,v,vec[1];

	vec[0]=arg;
	if(start)
	{
		start=0;
		init_perlin();
	}

	setup(0,bx0,bx1,rx0,rx1);

	sx=s_curve(rx0);

	u=rx0*g1[p[bx0]];
	v=rx1*g1[p[bx1]];

	return(lerp(sx,u,v));
}

double noise2(double vec[2])
{
	int bx0,bx1,by0,by1;
	double rx0,rx1,ry0,ry1,*q,sx,sy,a,b,t,u,v;
	int b0,b1;
	int b00,b10,b01,b11;

	if(start)
	{
		start=0;
		init_perlin();
	}

	setup(0,bx0,bx1,rx0,rx1);
	setup(1,by0,by1,ry0,ry1);

	b0=p[bx0];
	b1=p[bx1];

	b00=p[b0+by0];
	b10=p[b1+by0];
	b01=p[b0+by1];
	b11=p[b1+by1];

	sx=s_curve(rx0);
	sy=s_curve(ry0);

	q=g2[b00]; u=at2(rx0,ry0);
	q=g2[b10]; v=at2(rx1,ry0);
	a=lerp(sx,u,v);

	q=g2[b01]; u=at2(rx0,ry1);
	q=g2[b11]; v=at2(rx1,ry1);
	b=lerp(sx,u,v);

	return(lerp(sy,a,b));
}

double noise3(double vec[3])
{
	int bx0,bx1,by0,by1,bz0,bz1;
	double rx0,rx1,ry0,ry1,rz0,rz1,*q,sx,sy,sz,a,b,c,d,t,u,v;
	int b0,b1;
	int b00,b10,b01,b11;
	int b000,b100,b010,b110,b001,b101,b011,b111;

	if(start)
	{
		start=0;
		init_perlin();
	}

	setup(0,bx0,bx1,rx0,rx1);
	setup(1,by0,by1,ry0,ry1);
	setup(2,bz0,bz1,rz0,rz1);

	b0=p[bx0];
	b1=p[bx1];

	b00=p[b0+by0];
	b10=p[b1+by0];
	b01=p[b0+by1];
	b11=p[b1+by1];

	b000=p[b00+bz0];
	b100=p[b10+bz0];
	b010=p[b01+bz0];
	b110=p[b11+bz0];
	b001=p[b00+bz1];
	b101=p[b10+bz1];
	b011=p[b01+bz1];
	b111=p[b11+bz1];

	sx=s_curve(rx0);
	sy=s_curve(ry0);
	sz=s_curve(rz0);

	q=g3[b000]; u=at3(rx0,ry0,rz0);
	q=g3[b100]; v=at3(rx1,ry0,rz0);
	a=lerp(sx,u,v);

	q=g3[b010]; u=at3(rx0,ry1,rz0);
	q=g3[b110]; v=at3(rx1,ry1,rz0);
	b=lerp(sx,u,v);

	c=lerp(sy,a,b);

	q=g3[b001]; u=at3(rx0,ry0,rz1);
	q=g3[b101]; v=at3(rx1,ry0,rz1);
	a=lerp(sx,u,v);

	q=g3[b011]; u=at3(rx0,ry1,rz1);
	q=g3[b111]; v=at3(rx1,ry1,rz1);
	b=lerp(sx,u,v);

	d=lerp(sy,a,b);

	return(lerp(sz,c,d));
}

double noise4(double vec[4])
{
	int bx0,bx1,by0,by1,bz0,bz1,bt0,bt1;
	double rx0,rx1,ry0,ry1,rz0,rz1,rt0,rt1,*q,sx,sy,sz,st,a,b,c,d,e,f,t,u,v;
	int b0,b1;
	int b00,b10,b01,b11;
	int b000,b100,b010,b110,b001,b101,b011,b111;
	int b0000,b1000,b0100,b1100,b0010,b1010,b0110,b1110,
		b0001,b1001,b0101,b1101,b0011,b1011,b0111,b1111;

	if(start)
	{
		start=0;
		init_perlin();
	}

	setup(0,bx0,bx1,rx0,rx1);
	setup(1,by0,by1,ry0,ry1);
	setup(2,bz0,bz1,rz0,rz1);
	setup(3,bt0,bt1,rt0,rt1);

	b0=p[bx0];
	b1=p[bx1];

	b00=p[b0+by0];
	b10=p[b1+by0];
	b01=p[b0+by1];
	b11=p[b1+by1];

	b000=p[b00+bz0];
	b100=p[b10+bz0];
	b010=p[b01+bz0];
	b110=p[b11+bz0];
	b001=p[b00+bz1];
	b101=p[b10+bz1];
	b011=p[b01+bz1];
	b111=p[b11+bz1];

	b0000=p[b000+bt0];
	b1000=p[b100+bt0];
	b0100=p[b010+bt0];
	b1100=p[b110+bt0];
	b0010=p[b001+bt0];
	b1010=p[b101+bt0];
	b0110=p[b011+bt0];
	b1110=p[b111+bt0];
	b0001=p[b000+bt1];
	b1001=p[b100+bt1];
	b0101=p[b010+bt1];
	b1101=p[b110+bt1];
	b0011=p[b001+bt1];
	b1011=p[b101+bt1];
	b0111=p[b011+bt1];
	b1111=p[b111+bt1];

	sx=s_curve(rx0);
	sy=s_curve(ry0);
	sz=s_curve(rz0);
	st=s_curve(rt0);

	q=g4[b0000]; u=at4(rx0,ry0,rz0,rt0);
	q=g4[b1000]; v=at4(rx1,ry0,rz0,rt0);
	a=lerp(sx,u,v);

	q=g4[b0100]; u=at4(rx0,ry1,rz0,rt0);
	q=g4[b1100]; v=at4(rx1,ry1,rz0,rt0);
	b=lerp(sx,u,v);

	c=lerp(sy,a,b);

	q=g4[b0010]; u=at4(rx0,ry0,rz1,rt0);
	q=g4[b1010]; v=at4(rx1,ry0,rz1,rt0);
	a=lerp(sx,u,v);

	q=g4[b0110]; u=at4(rx0,ry1,rz1,rt0);
	q=g4[b1110]; v=at4(rx1,ry1,rz1,rt0);
	b=lerp(sx,u,v);

	d=lerp(sy,a,b);

	e=lerp(sz,c,d);



	q=g4[b0001]; u=at4(rx0,ry0,rz0,rt1);
	q=g4[b1001]; v=at4(rx1,ry0,rz0,rt1);
	a=lerp(sx,u,v);

	q=g4[b0101]; u=at4(rx0,ry1,rz0,rt1);
	q=g4[b1101]; v=at4(rx1,ry1,rz0,rt1);
	b=lerp(sx,u,v);

	c=lerp(sy,a,b);

	q=g4[b0011]; u=at4(rx0,ry0,rz1,rt1);
	q=g4[b1011]; v=at4(rx1,ry0,rz1,rt1);
	a=lerp(sx,u,v);

	q=g4[b0111]; u=at4(rx0,ry1,rz1,rt1);
	q=g4[b1111]; v=at4(rx1,ry1,rz1,rt1);
	b=lerp(sx,u,v);

	d=lerp(sy,a,b);

	f=lerp(sz,c,d);

	return(lerp(st,e,f));
}

static void normalize2(double v[2])
{
	double s=sqrt(v[0]*v[0]+v[1]*v[1]);
	v[0]/=s;
	v[1]/=s;
}

static void normalize3(double v[3])
{
	double s=sqrt(v[0]*v[0]+v[1]*v[1]+v[2]*v[2]);
	v[0]/=s;
	v[1]/=s;
	v[2]/=s;
}

static void normalize4(double v[4])
{
	double s=sqrt(v[0]*v[0]+v[1]*v[1]+v[2]*v[2]+v[3]*v[3]);
	v[0]/=s;
	v[1]/=s;
	v[2]/=s;
	v[3]/=s;
}



void init_perlin()
{
	int i,j,k;

	for(i=0;i<B;i++)
	{
		p[i]=i;

		g1[i]=(double)((grand()%(B+B))-B)/B;

		for(j=0;j<2;j++) g2[i][j]=(double)((grand()%(B+B))-B)/B;
		normalize2(g2[i]);

		for(j=0;j<3;j++) g3[i][j]=(double)((grand()%(B+B))-B)/B;
		normalize3(g3[i]);

		for(j=0;j<4;j++) g4[i][j]=(double)((grand()%(B+B))-B)/B;
		normalize4(g4[i]);
	}

	while(--i)
	{
		k=p[i];
		p[i]=p[j=grand()%B];
		p[j]=k;
	}

	for(i=0;i<B+2;i++)
	{
		p[B+i]=p[i];
		g1[B+i]=g1[i];
		for(j=0;j<2;j++) g2[B+i][j]=g2[i][j];
		for(j=0;j<3;j++) g3[B+i][j]=g3[i][j];
		for(j=0;j<4;j++) g4[B+i][j]=g4[i][j];
	}
}



/* --- My harmonic summing functions - PDB --------------------------*/

/*
   In what follows "alpha" is the weight when the sum is formed.
   Typically it is 2, As this approaches 1 the function is noisier.
   "beta" is the harmonic scaling/spacing, typically 2.
*/

double perlin_noise_1d(double x,double alpha,double beta,int n)
{
	int i;
	double val,sum=0;
	double p,scale=1;

	p=x;
	for(i=0;i<n;i++)
	{
		val=noise1(p);
		sum+=val/scale;
		scale*=alpha;
		p*=beta;
	}
	return(sum);
}

double perlin_noise_2d(double x,double y,double alpha,double beta,int n)
{
	int i;
	double val,sum=0;
	double p[2],scale=1;

	p[0]=x;
	p[1]=y;
	for(i=0;i<n;i++)
	{
		val=noise2(p);
		sum+=val/scale;
		scale*=alpha;
		p[0]*=beta;
		p[1]*=beta;
	}
	return(sum);
}

double perlin_noise_3d(double x,double y,double z,double alpha,double beta,int n)
{
	int i;
	double val,sum=0;
	double p[3],scale=1;

	p[0]=x;
	p[1]=y;
	p[2]=z;
	for(i=0;i<n;i++)
	{
		val=noise3(p);
		sum+=val/scale;
		scale*=alpha;
		p[0]*=beta;
		p[1]*=beta;
		p[2]*=beta;
	}
	return(sum);
}

double perlin_noise_4d(double x,double y,double z,double t,double alpha,double beta,int n)
{
	int i;
	double val,sum=0;
	double p[4],scale=1;

	p[0]=x;
	p[1]=y;
	p[2]=z;
	p[3]=t;
	for(i=0;i<n;i++)
	{
		val=noise4(p);
		sum+=val/scale;
		scale*=alpha;
		p[0]*=beta;
		p[1]*=beta;
		p[2]*=beta;
		p[3]*=beta;
	}
	return(sum);
}
