#import "LotsaSnowView.h"


@implementation LSLotsaSnowView

-(id)initWithFrame:(NSRect)frame isPreview:(BOOL)preview
{
	if(self=[super initWithFrame:frame isPreview:preview useGL:YES])
	{
		[self setAnimationTimeInterval:1/60.0];
		[self setConfigName:@"ConfigSheet"];
		[self setSaverDefaults:[NSDictionary dictionaryWithObjectsAndKeys:
			@"0.5",@"growthRate",
			[NSKeyedArchiver archivedDataWithRootObject:[NSColor colorWithDeviceRed:0.6 green:0.7 blue:1.0 alpha:1]],@"topCol",
			[NSKeyedArchiver archivedDataWithRootObject:[NSColor colorWithDeviceRed:0.2 green:0.3 blue:0.6 alpha:1]],@"bottomCol",
			[NSKeyedArchiver archivedDataWithRootObject:[NSColor colorWithDeviceRed:0.6 green:0.6 blue:0.6 alpha:1]],@"glareCol",
			@"0",@"useDesktop",
			@"0.25",@"desktopFade",
			@"0",@"useClock",
		nil]];

		flakes=[[NSMutableArray alloc] init];
		grower=new flake_grower(64);

		if(preview||[[self defaults] boolForKey:@"useDesktop"])
		screenshot=[[self grabScreenShotWithFrame:frame] retain];
		else screenshot=nil;
	}

    return self;
}

-(void)dealloc
{
	[flakes release];
	delete grower;

	[screenshot release];

	[super dealloc];
}

-(void)drawRect:(NSRect)aRect
{
	[screenshot drawAtPoint:NSMakePoint(0,0)];
}

-(void)startAnimationWithDefaults:(ScreenSaverDefaults *)defaults
{
	float rate=[defaults floatForKey:@"growthRate"];

	NSColor *toprgb=[[NSKeyedUnarchiver unarchiveObjectWithData:[defaults dataForKey:@"topCol"]] colorUsingColorSpaceName:NSCalibratedRGBColorSpace];
	NSColor *bottomrgb=[[NSKeyedUnarchiver unarchiveObjectWithData:[defaults dataForKey:@"bottomCol"]] colorUsingColorSpaceName:NSCalibratedRGBColorSpace];
	NSColor *glarergb=[[NSKeyedUnarchiver unarchiveObjectWithData:[defaults dataForKey:@"glareCol"]] colorUsingColorSpaceName:NSCalibratedRGBColorSpace];

	[toprgb getRed:&top_r green:&top_g blue:&top_b alpha:NULL];
	[bottomrgb getRed:&bottom_r green:&bottom_g blue:&bottom_b alpha:NULL];
	[glarergb getRed:&glare_r green:&glare_g blue:&glare_b alpha:NULL];

	[[self openGLContext] makeCurrentContext];

	t=0;
	max_growth=256;
	cycles_per_second=rate*rate*4000+100;

	flake_growth=0;
	growth_counter=0;
	screen_ratio=(float)screen_w/(float)screen_h;

	srand(time(0));

	glClearColor(0,0,0,0);
	glViewport(0,0,screen_w,screen_h);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glScalef(1/screen_ratio,1,1);

	if([[self defaults] boolForKey:@"useDesktop"])
	screentex=[LSGLConverter uncopiedTextureRectangleFromRep:screenshot];

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	[self animateOneFrame];
	[self resetDeltaTime];

	[NSOpenGLContext clearCurrentContext];
}

-(void)stopAnimation
{
	[[self openGLContext] makeCurrentContext];
	[flakes removeAllObjects];
	[NSOpenGLContext clearCurrentContext];

	[super stopAnimation];
}

-(void)animateOneFrame
{
	[[self openGLContext] makeCurrentContext];

	double dt=[self deltaTime];
	t+=dt;

	growth_counter+=cycles_per_second*dt;
	if(growth_counter>1000) growth_counter=1000;

	while(growth_counter>0)
	{
		grower->grow();

		flake_growth++;
		growth_counter--;

		if(flake_growth>=max_growth||grower->overgrown())
		{
			unsigned char *image=grower->make_image();
			LSSnowFlake *flake=[[[LSSnowFlake alloc] initWithData:image size:grower->size ratio:screen_ratio] autorelease];
			[flakes addObject:flake];
			[flakes sortUsingSelector:@selector(compare:)];

			grower->restart();
			flake_growth=0;
		}
	}

	glClear(GL_COLOR_BUFFER_BIT);

	glLoadIdentity();

	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);

//	qsort(flakes,max_flakes,sizeof(flake),flake::depthcmp);

	for(int i=0;i<(int)[flakes count];i++)
	{
		LSSnowFlake *flake=[flakes objectAtIndex:i];
		if([flake move:dt])
		{
			[flakes removeObjectAtIndex:i--];
		}
		else
		{
			[flake draw];
		}
	}

	glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_COLOR);

	if([[self defaults] boolForKey:@"useDesktop"])
	{
		glEnable(GL_TEXTURE_RECTANGLE_EXT);
		glBindTexture(GL_TEXTURE_RECTANGLE_EXT,screentex);

		glLoadIdentity();
		glScalef(screen_ratio,1,1);

		float tw=[screenshot pixelsWide];
		float th=[screenshot pixelsHigh];
		float fade=[[self defaults] floatForKey:@"desktopFade"];

		if(![self isPreview]&&t<1) fade=1-(1-fade)*(t*t*(3-2*t));

		glColor3f(fade,fade,fade);

		glBegin(GL_QUADS);
		glTexCoord2f(0,0); glVertex2f(-1,1);
		glTexCoord2f(tw,0); glVertex2f(1,1);
		glTexCoord2f(tw,th); glVertex2f(1,-1);
		glTexCoord2f(0,th); glVertex2f(-1,-1);
		glEnd();

		glDisable(GL_TEXTURE_RECTANGLE_EXT);
	}
	else
	{
		glDisable(GL_TEXTURE_2D);

		glLoadIdentity();
		glScalef(screen_ratio,1,1);

		glBegin(GL_QUADS);
		glColor3f(top_r,top_g,top_b);
		glVertex2f(-1,1);
		glVertex2f(1,1);
		glColor3f(bottom_r,bottom_g,bottom_b);
		glVertex2f(1,-1);
		glVertex2f(-1,-1);
		glEnd();

		glTranslatef(1,1,0);
		glScalef(-2,-2,1);

		glBegin(GL_TRIANGLE_FAN);

		glColor3f(glare_r,glare_g,glare_b);
		glVertex2f(0,0);
		glColor3f(0,0,0);

		int segments=16;
		for(int i=0;i<=segments;i++)
		{
			float a=float(i)/float(segments)*M_PI/2;
			glVertex2f(cos(a),sin(a));
		}

		glEnd();
	}

	glFlush();
	[[self openGLContext] flushBuffer];

	[NSOpenGLContext clearCurrentContext];
}

-(void)updateConfigWindow:(NSWindow *)window usingDefaults:(ScreenSaverDefaults *)defaults
{
	[growthrate setFloatValue:[defaults floatForKey:@"growthRate"]];

	[topcol setColor:(NSColor *)[NSKeyedUnarchiver unarchiveObjectWithData:[defaults dataForKey:@"topCol"]]];
	[bottomcol setColor:(NSColor *)[NSKeyedUnarchiver unarchiveObjectWithData:[defaults dataForKey:@"bottomCol"]]];
	[glarecol setColor:(NSColor *)[NSKeyedUnarchiver unarchiveObjectWithData:[defaults dataForKey:@"glareCol"]]];

	[usedesktop setIntValue:[defaults integerForKey:@"useDesktop"]];
	[desktopfade setFloatValue:[defaults floatForKey:@"desktopFade"]];

	[self makeNewSnowFlake:self];
}

-(void)updateDefaults:(ScreenSaverDefaults *)defaults usingConfigWindow:(NSWindow *)window
{
	[defaults setFloat:[growthrate floatValue] forKey:@"growthRate"];

	[defaults setObject:[NSKeyedArchiver archivedDataWithRootObject:[topcol color]] forKey:@"topCol"];
	[defaults setObject:[NSKeyedArchiver archivedDataWithRootObject:[bottomcol color]] forKey:@"bottomCol"];
	[defaults setObject:[NSKeyedArchiver archivedDataWithRootObject:[glarecol color]] forKey:@"glareCol"];

	[defaults setInteger:[usedesktop intValue] forKey:@"useDesktop"];
	[defaults setFloat:[desktopfade floatValue] forKey:@"desktopFade"];
}

-(IBAction)makeNewSnowFlake:(id)sender
{
	int size=grower->size;
	NSBitmapImageRep *rep=[[NSBitmapImageRep alloc] initWithBitmapDataPlanes:NULL
	pixelsWide:size*4 pixelsHigh:size*2 bitsPerSample:8 samplesPerPixel:4 hasAlpha:YES
	isPlanar:NO colorSpaceName:NSCalibratedRGBColorSpace bytesPerRow:0 bitsPerPixel:32];
	unsigned char *pixels=[rep bitmapData];
	int bytesperrow=[rep bytesPerRow];

	if(pixels)
	{
		grower->restart();

		for(int i=0;i<256;i++)
		{
			grower->grow();
			if(grower->overgrown()) break;
		}

		unsigned char *flakepixels=grower->make_image();

		for(int y=0;y<size*2;y++)
		{
			unsigned char *srcline=flakepixels+4*size*y;
			unsigned long *destline=(unsigned long *)(pixels+bytesperrow*y);
			for(int x=0;x<size*4;x++) *destline++=*srcline++*0x01010101;
		}

		NSSize size=[flakeview frame].size;
		size.width-=4;
		size.height-=4;
		NSImage *image=[[[NSImage alloc] initWithSize:size] autorelease];

		[image lockFocus];
		[[NSGraphicsContext currentContext] setImageInterpolation:NSImageInterpolationHigh];
		[rep drawInRect:NSMakeRect(0,0,size.width,size.height)];
		[flakeview setImage:image];
		[image unlockFocus];
	}
}

+(BOOL)performGammaFade
{
	return ![[ScreenSaverDefaults defaultsForModuleWithName:[[NSBundle bundleForClass:self] bundleIdentifier]] boolForKey:@"useDesktop"];
}


@end



@implementation LSSnowFlake

-(id)initWithData:(void *)data size:(int)datasize ratio:(float)ratio;
{
	if(self=[super init])
	{
		screen_ratio=ratio;

		z=drand();
		size=0.13/(1+z);

		y0=1+size;
		x0=(2*drand()-1)*(screen_ratio+size);

		a=drand()*2*M_PI;
		dadt=(2*drand()-1)*0.3;

		t=0;

		glGenTextures(1,&tex);
		glBindTexture(GL_TEXTURE_2D,tex);

		glPixelStorei(GL_UNPACK_ALIGNMENT,1);

		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
//		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_NEAREST);
		glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);

//		glTexImage2D(GL_TEXTURE_2D,0,GL_ALPHA,4*SIZE,2*SIZE,
//		0,GL_ALPHA,GL_UNSIGNED_BYTE,texture);
		gluBuild2DMipmaps(GL_TEXTURE_2D,GL_ALPHA,4*datasize,2*datasize,
		GL_ALPHA,GL_UNSIGNED_BYTE,data);
	}

	return self;
}

-(void)dealloc
{
	glDeleteTextures(1,&tex);

	[super dealloc];
}

-(NSComparisonResult)compare:(LSSnowFlake *)flake
{
	return z>[flake z]?NSOrderedAscending:NSOrderedDescending;
}

-(BOOL)move:(double)dt
{
	t+=dt;

//	x+=0.03*(2*drand()-1)*dt;
//	y-=0.1*dt/(z+1);

//	y=y0-(0.1*t-0.01*sin(2*t))/(z+1);
//	x=x0+(cos(t-PI*0.33)*0.02)/(z+1);

	y=y0-0.1*t/(z+1);
	x=x0;
//	x=x0+0.02*t*(2*z-1)/(z+1);

	a+=dadt*dt;

	if(y<-1-size) return YES;
//	if(x>screen_ratio+size) x-=2*(screen_ratio+size);
//	if(x<-screen_ratio-size) x+=2*(screen_ratio+size);

	return NO;
}

-(void)draw
{
	glBindTexture(GL_TEXTURE_2D,tex);

	glColor3f(1-z,1-z,1-z);

	glPushMatrix();
	glTranslatef(x,y,0);
	glRotatef(a/M_PI*180,0,0,1);
	glScalef(size,size/1.15470054,1);

	glBegin(GL_QUADS);
	glTexCoord2f(1,0); glVertex2f(1,1);
	glTexCoord2f(1,1); glVertex2f(1,-1);
	glTexCoord2f(0,1); glVertex2f(-1,-1);
	glTexCoord2f(0,0); glVertex2f(-1,1);
	glEnd();

	glPopMatrix();
}

-(float)z { return z; }

@end



@implementation LSShadowedImageView

-(void)drawRect:(NSRect)rect
{
	[NSGraphicsContext saveGraphicsState];
	NSShadow *shadow=[[NSShadow alloc] init];

	[shadow setShadowBlurRadius:4];
	[shadow setShadowColor:[NSColor blackColor]];
	[shadow set];

	[super drawRect:rect];

	[NSGraphicsContext restoreGraphicsState];
	[shadow release];
}

@end
