#import <math.h>
#import <stdlib.h>

#include "snow.h"



// flake_grower

flake_grower::flake_grower(int size)
{
	this->size=size;
	buf_size=((size/2)*(size/2+1));

	water1=new float[buf_size];
	water2=new float[buf_size];
	ice=new float[buf_size];
	frozen=new int[buf_size];
	neighbours=new int[buf_size];
	texture=new unsigned char[size*size*8];

	for(int i=0;i<size*size*8;i++) texture[i]=0;

	restart();
}

flake_grower::~flake_grower()
{
	delete[] water1;
	delete[] water2;
	delete[] ice;
	delete[] frozen;
	delete[] neighbours;
	delete[] texture;
}

void flake_grower::restart()
{
	water=water1;
	oldwater=water2;

	for(int i=0;i<buf_size;i++)
	{
		water[i]=external;
		ice[i]=0;
		frozen[i]=0;
	}

	frozen[buf_size-1]=1;
	water[buf_size-1]=0;
//	frozen[buf_size/2]=1;

	icelimit=2;
	p=1.0/6.0;
	external=1;
	external_rate=0.01;

	t_grow=0;
}

void inline flake_grower::find_neighbours_even(int x,int y,int w,int i,int &c1,int &c2,int &c3,int &c4,int &c5,int &c6)
{
	if(y==0) // top row
	{
		if(x==0)
		{
			c1=c3=c4=1;
			c2=c5=c6=w;
		}
		else if(x==w-1)
		{
			c1=i-1;
			c2=i+w; c3=c2-1;
			c4=i;
			c5=c2; c6=c3;
		}
		else
		{
			c1=i-1; c2=i+1;
			c3=i+w; c4=c3-1;
			c5=c3; c6=c4;
		}
	}
	else if(y==size/2-1) // next-to-last cell
	{
		c1=i+1; c2=i-1;
		c3=i-2;
		c4=c5=i; c6=c2;
	}
	else // mid rows
	{
		if(x==0)
		{
			c1=i-w; c2=c1-1;
			c3=i+1; c4=i+w;
			c5=c1; c6=c3;
		}
		else if(x==w-1)
		{
			c1=i-w; c2=c1-1;
			c3=i-1; c4=i;
			c5=i+w; c6=c5-1;
		}
		else
		{
			c1=i-w; c2=c1-1;
			c3=i-1; c4=i+1;
			c5=i+w; c6=c5-1;
		}
	}
}

void inline flake_grower::find_neighbours_odd(int x,int y,int w,int i,int &c1,int &c2,int &c3,int &c4,int &c5,int &c6)
{
	if(x==0)
	{
		c1=i-w; c2=c1+1;
		c3=i+1; c4=i+w;
		c5=c2; c6=c3;
	}
	else if(x==w-1)
	{
		c1=i-w; c2=i-1;
		c3=i+w-1;
		c4=c1; c5=c2; c6=c3;
	}
	else
	{
		c1=i-w; c2=c1+1;
		c3=i-1; c4=i+1;
		c5=i+w; c6=c5-1;
	}
}

void flake_grower::grow()
{
	int i;

	icelimit*=(0.96+0.08*drand());
	external*=(0.96+0.08*drand());

//	if(p>1.0/6.0) p=1.0/6.0;
//	if(external_rate>1) external_rate=1;

	float *wtmp=water;
	water=oldwater;
	oldwater=wtmp;

	i=0;
	for(int y=0;y<size/2;y++)
	{
		int width=size/2-y;

		for(int x=0;x<width;x++)
		{
			if(!frozen[i])
			{
				int c1,c2,c3,c4,c5,c6;
				find_neighbours_even(x,y,width,i,c1,c2,c3,c4,c5,c6);

				int n=0;
				if(frozen[c1]) n++; if(frozen[c2]) n++;
				if(frozen[c3]) n++; if(frozen[c4]) n++;
				if(frozen[c5]) n++; if(frozen[c6]) n++;
				neighbours[i]=n;

				if(n) { ice[i]+=oldwater[i]; oldwater[i]=0; }
			}

			i++;
		}

		for(int x=0;x<width;x++)
		{
			if(!frozen[i])
			{
				int c1,c2,c3,c4,c5,c6;
				find_neighbours_odd(x,y,width,i,c1,c2,c3,c4,c5,c6);

				int n=0;
				if(frozen[c1]) n++; if(frozen[c2]) n++;
				if(frozen[c3]) n++; if(frozen[c4]) n++;
				if(frozen[c5]) n++; if(frozen[c6]) n++;
				neighbours[i]=n;

				if(n) { ice[i]+=oldwater[i]; oldwater[i]=0; }
			}

			i++;
		}
	}

	i=0;
	for(int y=0;y<size/2;y++)
	{
		int width=size/2-y;

		for(int x=0;x<width;x++)
		{
			if(!frozen[i])
			{
				int c1,c2,c3,c4,c5,c6;
				find_neighbours_even(x,y,width,i,c1,c2,c3,c4,c5,c6);

				float in=0;
				in+=oldwater[c1]; in+=oldwater[c2];
				in+=oldwater[c3]; in+=oldwater[c4];
				in+=oldwater[c5]; in+=oldwater[c6];

				if(!neighbours[i]) // diffuse water + external inflow;
				{
					float w=oldwater[i]*(1-p*6)+in*p;
					water[i]=w+(external-w)*external_rate;
				}
				else // freeze + external inflow
				{
					ice[i]+=in*p+external*external_rate;
					water[i]=0;

					if(ice[i]>icelimit) frozen[i]=t_grow;
				}
			}

			i++;
		}

		for(int x=0;x<width;x++)
		{
			if(!frozen[i])
			{
				int c1,c2,c3,c4,c5,c6;
				find_neighbours_odd(x,y,width,i,c1,c2,c3,c4,c5,c6);

				float in=0;
				in+=oldwater[c1]; in+=oldwater[c2];
				in+=oldwater[c3]; in+=oldwater[c4];
				in+=oldwater[c5]; in+=oldwater[c6];

				if(!neighbours[i]) // diffuse water + external inflow;
				{
					float w=oldwater[i]*(1-p*6)+in*p;
					water[i]=w+(external-w)*external_rate;
				}
				else // freeze + external inflow
				{
					ice[i]+=in*p+external*external_rate;
					water[i]=0;

					if(ice[i]>icelimit) frozen[i]=t_grow;
				}
			}

			i++;
		}
	}

	t_grow++;
}

bool flake_grower::overgrown()
{
	for(int x=0;x<size/2;x++)
	if(ice[x]) return(true);

	return(false);
}

unsigned char *flake_grower::make_image()
{
	int i=0;

	for(int y=0;y<size;y++)
	{
		int width=size/2-y/2;
		int yy=size-y-1;

		for(int x=0;x<width;x++)
		{
			int col;

			if(frozen[i])
			{
				col=0xff;
			}
			else
			{
				col=0x00;
			}

			put_hex(yy,-x,0,col);
			put_hex(yy,0,x,col);
			put_hex(-x,yy,0,col);
			put_hex(0,yy,-x,col);
			put_hex(x,0,yy,col);
			put_hex(0,-x,yy,col);

			put_hex(-yy,x,0,col);
			put_hex(-yy,0,-x,col);
			put_hex(x,-yy,0,col);
			put_hex(0,-yy,x,col);
			put_hex(-x,0,-yy,col);
			put_hex(0,x,-yy,col);

			i++;
		}
	}

	return texture;
}

void inline flake_grower::put_hex(int u,int v,int w,int c)
{
	int offs=
		4*size*size-2*size-2
		+2*u+v-w
		+(v+w)*4*size;

	texture[offs]=c;
	texture[offs+1]=c;
}

