#include "water.h"

#if (defined(__ppc__)||defined(__ppc64__))

#define ALTIVEC_AVAILABLE

#elif defined(__i386__)

#define SSE_AVAILABLE

#endif




//const complex cerf(const complex &z);



// water

water::water(int w,int h,float v,float b,float lx,float ly):state(w,h)
{
	init(w,h,w,h,v,b,lx,ly);
}

water::water(int w,int h,int max_p,int max_q,float v,float b,float lx,float ly):state(max_p,max_q)
{
	init(w,h,max_p,max_q,v,b,lx,ly);
}

water::~water()
{
	delete wpq;
	delete ampl;
	delete sin_px;
	delete cos_px;
	delete sin_qy;
	delete cos_qy;
	delete z;
	delete n;
}

void water::init(int w,int h,int max_p,int max_q,float v,float b,float lx,float ly)
{
	this->w=w;
	this->h=h;
	this->max_p=max_p;
	this->max_q=max_q;

	t0=0;

	wpq=new float[max_p*max_q];
	ampl=new float[max_p*max_q];
	sin_px=new float[w*max_p];
	cos_px=new float[w*max_p];
	sin_qy=new float[h*max_q];
	cos_qy=new float[h*max_q];
	z=new float[w*h];
	n=new vector[w*h];

	this->v=v;
	this->b=b;
	this->lx=lx;
	this->ly=ly;

	int i;

	i=0;
	for(int q=0;q<max_q;q++)
	for(int p=0;p<max_p;p++)
	{
		float discr=(float(p))*(float(p))/(lx*lx)+(float(q))*(float(q))/(ly*ly)-b*b*v*v/(4*M_PI*M_PI); //-b*b*v*v/(2*M_PI*M_PI)

		if(discr>0) wpq[i]=v*M_PI*sqrt(discr);
		else wpq[i]=-v*M_PI*sqrt(-discr);
		i++;
	}

	i=0;
	for(int x=0;x<w;x++)
	for(int p=0;p<max_p;p++)
	{
		sin_px[i]=float(p)*M_PI/lx*sin(float(p)*M_PI*float(x)/float(w-1));
		cos_px[i]=cos(float(p)*M_PI*float(x)/float(w-1));
		i++;
	}

	i=0;
	for(int y=0;y<h;y++)
	for(int q=0;q<max_q;q++)
	{
		sin_qy[i]=float(q)*M_PI/ly*sin(float(q)*M_PI*float(y)/float(h-1));
		cos_qy[i]=cos(float(q)*M_PI*float(y)/float(h-1));
		i++;
	}

	i=0;
	for(int y=0;y<h;y++)
	for(int x=0;x<w;x++)
	{
		n[i].z=1;
		i++;
	}
}


void water::calculate(float t)
{
	int i;

	float decay=exp(-v*v*b*(t-t0)/2);

	i=0;
	for(int q=0;q<max_q;q++)
	for(int p=0;p<max_p;p++)
	{
		if(wpq[i]>=0)
		ampl[i]=(state.apq[i]*cos(wpq[i]*(t-t0))+state.bpq[i]*sin(wpq[i]*(t-t0)))*decay;
		else
		ampl[i]=state.apq[i]*exp(wpq[i]*(t-t0));

		i++;
	}

	i=0;
	for(int y=0;y<h;y++)
	for(int x=0;x<w;x++)
	{
		float z_curr=0;
		float ndzdx=0;
		float ndzdy=0;
		float *sin_px_ptr_start=&sin_px[x*max_p];
		float *cos_px_ptr_start=&cos_px[x*max_p];
		float *sin_qy_ptr=&sin_qy[y*max_q];
		float *cos_qy_ptr=&cos_qy[y*max_q];
		float *ampl_ptr=ampl;

		for(int q=0;q<max_q;q++)
		{
#ifdef ALTIVEC_AVAILABLE
			__vector float *v_sin_px=(__vector float *)sin_px_ptr_start;
			__vector float *v_cos_px=(__vector float *)cos_px_ptr_start;
			__vector float *v_ampl=(__vector float *)ampl_ptr;
			__vector float v_ampl_sin={0},v_ampl_cos={0};

			for(int p=0;p<max_p/4;p++)
			{
				v_ampl_sin=vec_madd(v_sin_px[p],v_ampl[p],v_ampl_sin);
				v_ampl_cos=vec_madd(v_cos_px[p],v_ampl[p],v_ampl_cos);
			}

			float *f;
			f=(float *)&v_ampl_sin;
			float ampl_sin=f[0]+f[1]+f[2]+f[3];
			f=(float *)&v_ampl_cos;
			float ampl_cos=f[0]+f[1]+f[2]+f[3];

			ampl_ptr+=max_p;
#else
			float *sin_px_ptr=sin_px_ptr_start;
			float *cos_px_ptr=cos_px_ptr_start;
			float ampl_cos=0;
			float ampl_sin=0;

			for(int p=0;p<max_p;p++)
			{
				float ampl=*ampl_ptr++;
				ampl_sin+=ampl*(*sin_px_ptr++);
				ampl_cos+=ampl*(*cos_px_ptr++);
			}
#endif

			ndzdx+=ampl_sin*(*cos_qy_ptr);
			ndzdy+=ampl_cos*(*sin_qy_ptr++);
			z_curr+=ampl_cos*(*cos_qy_ptr++);
		}

		z[i]=z_curr;
		n[i].x=ndzdx;
		n[i].y=ndzdy;

		i++;
	}

/*	i=0;
	for(int y=0;y<h;y++)
	for(int x=0;x<w;x++)
	{
		float nx,ny;

		if(x==0||x==w-1) n[i].x=0;
		else n[i].x=(z[i-1]-z[i+1])*float(w-1);

		if(y==0||y==h-1) n[i].y=0;
		else n[i].y=(z[i-w]-z[i+w])*float(h-1);

		n[i].z=2.0;

		i++;
	}*/
}

void water::add_state(water_state &other,float t)
{
	float decay=exp(-v*v*b*(t-t0)/2);

	int i=0;
	for(int q=0;q<max_q;q++)
	for(int p=0;p<max_p;p++)
	{
		float napq,nbpq;

		if(wpq[i]>0)
		{
			napq=state.apq[i]*cos(wpq[i]*(t-t0))+state.bpq[i]*sin(wpq[i]*(t-t0));
			nbpq=-state.apq[i]*sin(wpq[i]*(t-t0))+state.bpq[i]*cos(wpq[i]*(t-t0));
		}
		else
		{
			napq=state.apq[i]*exp(wpq[i]*(t-t0));
			nbpq=0;
		}

		state.apq[i]=napq*decay+other.apq[i];
		state.bpq[i]=nbpq*decay+other.bpq[i];
		i++;
	}

	t0=t;
}



// water_state

water_state::water_state(int max_p,int max_q)
{
	this->max_p=max_p;
	this->max_q=max_q;

	apq=new float[max_p*max_q];
	bpq=new float[max_p*max_q];

	int i=0;
	for(int q=0;q<max_q;q++)
	for(int p=0;p<max_p;p++)
	{
		apq[i]=bpq[i]=0;
		i++;
	}
}

water_state::~water_state()
{
	delete apq;
	delete bpq;
}


// random_state

random_state::random_state(water &water):water_state(water.max_p,water.max_q)
{
	int i=0;
	for(int q=0;q<max_q;q++)
	for(int p=0;p<max_p;p++)
	{
		if(water.wpq[i]>0)
		{
			apq[i]=0.03*(2*drand()-1)*exp(-0.01*water.wpq[i]*water.wpq[i]); // boltzmann distribution
			bpq[i]=0.03*(2*drand()-1)*exp(-0.01*water.wpq[i]*water.wpq[i]); // should fix the energy
		}
		else apq[i]=bpq[i]=0;

		i++;
	}
}

// drip_state

/*drip_state::drip_state(float x0,float y0,water &water):water_state(water.max_p,water.max_q)
{
	int i=0;
	for(int q=0;q<max_q;q++)
	for(int p=0;p<max_p;p++)
	{
		apq[i]=0;
		bpq[i]=0.003*4/(water.wpq[i]*water.lx*water.ly)*
		(cos(float(p)*M_PI*x0/water.lx)*cos(float(q)*M_PI*y0/water.ly));

		if(p==0) bpq[i]/=2;
		if(q==0) bpq[i]/=2;

		i++;
	}
}*/

// drip_state

drip_state::drip_state(float x0,float y0,float d,float ampl,water &water):water_state(water.max_p,water.max_q)
{
	int i=0;
	for(int q=0;q<max_q;q++)
	for(int p=0;p<max_p;p++)
	{
		apq[i]=0;

		bpq[i]=ampl*4/(water.wpq[i]*water.lx*water.ly)
		*cos(float(p)*M_PI*x0/water.lx)*cos(float(q)*M_PI*y0/water.ly)
		*exp(-M_PI*M_PI*d*d*(float(p*p)/(water.lx*water.lx)+float(q*q)/(water.ly*water.ly))/4)
/*		*0.5*(
			cerf(complex((water.lx-x0)/d,float(p)*M_PI*d/(2*water.lx))).re
			-cerf(complex(-x0/d,float(p)*M_PI*d/(2*water.lx))).re
		)
		*0.5*(
			cerf(complex((water.ly-y0)/d,float(q)*M_PI*d/(2*water.ly))).re
			-cerf(complex(-y0/d,float(q)*M_PI*d/(2*water.ly))).re
		)*/;


		if(p==0) bpq[i]/=2;
		if(q==0) bpq[i]/=2;

		i++;
	}
}

